/*
  VFB-01 / Virtual FB-01 module

  Copyright 2000 by Daisuke Nagano <breeze.nagano@nifty.ne.jp>
  Feb.10.2000


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* ------------------------------------------------------------------- */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>

#ifdef STDC_HEADERS
# include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
#endif

#include <getopt.h>

#include "vfb01.h"
#include "gettext_wrapper.h"
#include "version.h"

/* ------------------------------------------------------------------- */

#define PATH_BUF_SIZE 1024

static void  error_end( char * );
static int   option_gets( int, char ** );
static void  usage( void );
static void  display_version( void );

static char *command_name=NULL;
static int   verbose=FLAG_FALSE;
static char *dsp_device=NULL;
static char *midi_device=NULL;
static int   master_volume=127;
static int   units=VFB_MAX_CHANNEL_NUMBER;
static char *voice_parameter_file;
static int   is_use_fragment=FLAG_TRUE;
static int   is_output_to_stdout=FLAG_FALSE;
static int   is_ignore_ch10=FLAG_FALSE;

static VFB_DATA *vfb=NULL;

/* ------------------------------------------------------------------- */

int main( int argc, char **argv ) {

  int num;

#ifdef ENABLE_NLS
  setlocale( LC_ALL, "" );
  bindtextdomain( PACKAGE, LOCALEDIR );
  textdomain( PACKAGE );
#endif /* ENABLE_NLS */

  num = option_gets( argc, argv );

  /* open devices */

  if (openVirtualMidiDevice(midi_device)) {
    fprintf( stderr, _("%s: Cannot open device: %s\n"), command_name, midi_device );
    exit(1);
  }

  /* main loop */

  vfb = (VFB_DATA *)malloc(sizeof(VFB_DATA));
  if ( vfb == NULL ) {
    perror(command_name);
    exit(1);
  }

  vfb->dsp_device           = dsp_device;
  vfb->midi_device          = midi_device;
  vfb->is_use_fragment      = is_use_fragment;
  vfb->is_output_to_stdout  = is_output_to_stdout;
  vfb->is_normal_exit       = FLAG_TRUE;
  vfb->is_ignore_ch10       = is_ignore_ch10;
  vfb->verbose              = verbose;

  vfb->voice_parameter_file = voice_parameter_file;
  vfb->master_volume        = master_volume;
  vfb->units                = units;

  vfb01( vfb );

  /* finished */

  closeVirtualMidiDevice();

  exit(0);
}

/* ------------------------------------------------------------------- */

void error_end( char *msg ) {

  fprintf( stderr, "%s: %s\n", command_name, msg );
  exit(1);
}

static int option_gets( int argc, char **argv ) {

  extern char *optarg;
  extern int optind;

  int c;
  int option_index=0;

  command_name = 
    (strrchr(argv[0],'/')==NULL) ? argv[0] : (strrchr(argv[0],'/')+1);

  verbose              = FLAG_FALSE;
  midi_device          = MIDI_DEVICE_NAME;
  dsp_device           = DSP_DEVICE_NAME;
  voice_parameter_file = VOICE_PARAMETER_NAME;
  is_use_fragment      = FLAG_TRUE;
  is_output_to_stdout  = FLAG_FALSE;
  is_ignore_ch10       = FLAG_FALSE;

  while(1) {
    static struct option long_options[] = {
      {"dspdev",           1, 0, 'd'},
      {"mididev",          1, 0, 'm'},
      {"master-volume",    1, 0, 'a'},
      {"units",            1, 0, 'u'},
      {"voice-parameter",  1, 0, 't'},
      {"version",          0, 0, 'V'},
      {"verbose",          0, 0, 'v'},
      {"disable-fragment", 0, 0, 'A'},
      {"output-to-stdout", 0, 0, 's'},
      {"ignore-ch10",      0, 0, '0'},
      {"help",             0, 0, 'h'},

      {0, 0, 0, 0}
    };

    c = getopt_long(argc, argv, "AVvd:u:s:m:a:t:h", long_options, &option_index );
    if ( c == EOF ) break;

    switch(c) {

    case 'h': /* help */
      usage();
      break;

    case '0':
      is_ignore_ch10 = FLAG_TRUE;
      break;

    case 'A':
      is_use_fragment = FLAG_FALSE;
      break;

    case 's':
      is_output_to_stdout = FLAG_FALSE;
      break;

    case 'V':
      display_version(); /* version */
      break;

    case 'a': /* master volume */
      master_volume = atoi(optarg);
      if ( master_volume < 0 ) master_volume = 0;
      break;

    case 'u': /* units */
      units = atoi(optarg);
      if ( units < 0 ) units = 0;
      if ( units > VFB_MAX_CHANNEL_NUMBER ) units = VFB_MAX_CHANNEL_NUMBER;
      break;

    case 't': /* voice parameters */
      voice_parameter_file = strdup(optarg);
      break;

    case 'd':
      dsp_device = strdup(optarg);
      break;

    case 'm':
      midi_device = strdup(optarg);
      break;

    case 'v':
      verbose = FLAG_TRUE;
      break;

    case '?':
      break;

    default:
      break;
    }
  }

  if ( optind > argc ) {
    fprintf(stderr, _("%s: Invalid option specified.\n"), command_name);
    exit(1);
  } 

  return optind;
}

/* ------------------------------------------------------------------- */

static void usage( void ) {

  fprintf(stderr, "usage: %s [options]\n", command_name );
  fprintf(stderr, _("Options:\n"));

  fprintf(stderr, " -d,     --dspdev <name>         ");
  fprintf(stderr, _("DSP device name.\n"));
  fprintf(stderr, " -m,     --mididev <name>        ");
  fprintf(stderr, _("Raw MIDI device name.\n"));
  fprintf(stderr, " -a,     --master-volume <val>   ");
  fprintf(stderr, _("Master volume (\%).\n"));
  fprintf(stderr, " -u,     --units <val>           ");
  fprintf(stderr, _("Number of VFB-01.\n"));
  fprintf(stderr, " -t,     --voice-parameter <path>");
  fprintf(stderr, _("Tone parameter table.\n"));
  fprintf(stderr, " -A,     --disable-fragment      ");
  fprintf(stderr, _("Disable fragment of dsp device.\n"));
  fprintf(stderr, " -V,     --version               ");
  fprintf(stderr, _("Show version information.\n"));
  fprintf(stderr, " -h,     --help                  ");
  fprintf(stderr, _("Show this help message.\n"));

  exit(0);
}

static void display_version( void ) {

  fprintf(stderr, "%s version ", command_name );
  fprintf(stderr, VERSION_ID "\n");
  fprintf(stderr, "Virtual FB-01\n");

  fprintf(stderr, "Copyright 2000 by NAGANO Daisuke <breeze.nagano@nifty.ne.jp>\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "This is free software; see the source for copying conditions.\n");
  fprintf(stderr, "There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A\n");
  fprintf(stderr, "PARTICULAR PURPOSE.\n\n");
  fprintf(stderr, "Software implementation of YM-2151\n");
  fprintf(stderr, "(C) 1997,1998,1999 Jarek Burczynski (s0246@priv4.onet.pl)\n\n");

  exit(0);
}
