/*
  VFB-01 : Virtual FB-01 emulator

  Copyright 2000 by Daisuke Nagano <breeze.nagano@nifty.ne.jp>
  Feb.10.2000

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* ------------------------------------------------------------------- */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>

#include <signal.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>

#ifdef _POSIX_PRIORITY_SCHEDULING
# include <sched.h>
#endif
#ifdef _POSIX_MEMLOCK
# include <sys/mman.h>
#endif

#include "vfb01.h"
#include "smf.h"
#include "gettext_wrapper.h"

/* ------------------------------------------------------------------- */

#define VFB_DEBUG 1
#undef  VFB_DEBUG

/* ------------------------------------------------------------------- */

static int note_off( MidiEvent * );
static int note_on( MidiEvent * );
static int key_pressure( MidiEvent * );
static int program_change( MidiEvent * );
static int channel_pressure( MidiEvent * );
static int pitch_wheel( MidiEvent * );
static int control_change( MidiEvent * );
static int system_exclusive( MidiEvent * );

static void set_signals( void );
static void priority_init( void );

/* ------------------------------------------------------------------- */

static unsigned char sysex_buf[SYSEX_BUF_SIZE];
static VFB_DATA *evfb=NULL;
static int rpn_adr[VFB_MAX_CHANNEL_NUMBER];
static int nrpn_adr[VFB_MAX_CHANNEL_NUMBER];

/* ------------------------------------------------------------------- */

int vfb01( VFB_DATA *vfb ) {

  MidiEvent e;
  int i;

  if ( evfb == NULL && vfb != NULL ) evfb=vfb;
  if ( pcm8_open( vfb ) ) return 1;

  if ( setup_voices( vfb ) ) return 1;
  if ( setup_ym2151( vfb ) ) return 1;

  set_signals();
  priority_init();

  if ( evfb->is_output_to_stdout == FLAG_FALSE ) 
    pcm8_start();

  vfb->elapsed_time = 0;
  vfb->total_count = 0;


  for ( i=0 ; i<VFB_MAX_CHANNEL_NUMBER ; i++ ) {
    rpn_adr[i]  = 0xffff;
    nrpn_adr[i] = 0xffff;
  }


  /* main loop */

  while (1) {
    if ( !getMidiEvent( &e ) ) {
#ifdef VFB_DEBUG
      fprintf(stderr,_("Invalid data\n"));
#endif
      continue;
    }

    if ( e.ch >= vfb->units ) continue;
    if ( vfb->is_ignore_ch10 == FLAG_TRUE &&
	 e.ch == 9 ) continue;

    switch ( e.type ) {
    case MIDI_NOTEOFF:
      note_off( &e );
      break;

    case MIDI_NOTEON:
      note_on( &e );
      break;

    case MIDI_PRESSURE:
      key_pressure( &e );
      break;

    case MIDI_CONTROL:
      control_change( &e );
      break;

    case MIDI_PROGRAM:
      program_change( &e );
      break;

    case MIDI_CHANPRES:
      channel_pressure( &e );
      break;

    case MIDI_PITCHB:
      pitch_wheel( &e );
      break;

    case MIDI_SYSEX:
      system_exclusive( &e );
      break;

    default:
      break;
    }
    for ( i=0 ; i<vfb->units ; i++ ) {
      ym2151_set_freq_volume(i);
    }
  }

  /* finalize all resources */

#ifdef _POSIX_PRIORITY_SCHEDULING
  sched_yield();
#endif
#ifdef _POSIX_MEMLOCK
  munlockall();
#endif

  return 0;
}

/* ------------------------------------------------------------------- */

static int note_off( MidiEvent *ev ) {

#ifdef VFB_DEBUG
  fprintf(stdout, "NOTEOFF:  %02x %02x\n", ev->ch, ev->a);
#endif

  ym2151_note_off( ev->ch, ev->a );

  return 0;
}

static int note_on( MidiEvent *ev ) {

#ifdef VFB_DEBUG
  fprintf(stdout, "NOTE_ON:  %02x %02x %02x\n", ev->ch, ev->a, ev->b);
#endif

  if ( ev->b > 0 ) 
    ym2151_note_on( ev->ch, ev->a, ev->b );
  else
    ym2151_note_off( ev->ch, ev->a );

  return 0;
}

static int key_pressure( MidiEvent *ev ) {

#ifdef VFB_DEBUG
  fprintf(stdout, "KEYPRES:  %02x %02x %02x\n", ev->ch, ev->a, ev->b);
#endif
  return 0;
}

static int program_change( MidiEvent *ev ) {

#ifdef VFB_DEBUG
  fprintf(stdout, "PROGCHG:  %02x %02x\n", ev->ch, ev->a);
#endif

  ym2151_set_voice( ev->ch, ev->a );

  return 0;
}

static int channel_pressure( MidiEvent *ev ) {

#ifdef VFB_DEBUG
  fprintf(stdout, "CHPRESS:  %02x %02x\n", ev->ch, ev->a);
#endif
  return 0;
}

static int pitch_wheel( MidiEvent *ev ) {


  ym2151_set_bend( ev->ch, (ev->b<<7)+ev->a );

#ifdef VFB_DEBUG
  fprintf(stdout, "PITCH:    %02x %04d\n", ev->ch, (ev->b<<7)+ev->a);
#endif
  return 0;
}

/* ------------------------------------------------------------------- */

static int control_change( MidiEvent *ev ) {

  int i;

  switch( ev->a ) {
  case SMF_CTRL_BANK_SELECT_M:
    break;

  case SMF_CTRL_MODULATION_DEPTH:
    ym2151_set_modulation_depth( ev->ch, ev->b );
    break;

  case SMF_CTRL_BLESS_TYPE:
    break;

  case SMF_CTRL_FOOT_TYPE:
    break;

  case SMF_CTRL_PORTAMENT_TIME:
    ym2151_set_portament( ev->ch, ev->b );
    break;

  case SMF_CTRL_DATA_ENTRY_M:
    if ( rpn_adr[ev->ch] == 0x0000 ) {
      /* pitch bend sensitivity */
      ym2151_set_bend_sensitivity( ev->ch, ev->b, -1 );
    }
    break;

  case SMF_CTRL_MAIN_VOLUME:
    ym2151_set_master_volume( ev->ch, ev->b );
    break;

  case SMF_CTRL_BALLANCE_CTRL:
    break;

  case SMF_CTRL_PANPOT:
    pcm8_pan( ev->ch, ev->b );
    break;

  case SMF_CTRL_EXPRESSION:
    ym2151_set_expression( ev->ch, ev->b );
    break;


  case SMF_CTRL_BANK_SELECT_L:
    break;

  case SMF_CTRL_DATA_ENTRY_L:
    if ( rpn_adr[ev->ch] == 0x0000 ) {
      /* pitch bend sensitivity */
      ym2151_set_bend_sensitivity( ev->ch, -1, ev->b );
    }
    break;


  case SMF_CTRL_HOLD1:
    ym2151_set_hold( ev->ch,  ev->b>64?FLAG_TRUE:FLAG_FALSE );
    break;

  case SMF_CTRL_PORTAMENT:
    ym2151_set_portament_on( ev->ch, ev->b>64?FLAG_TRUE:FLAG_FALSE );
    break;

  case SMF_CTRL_SUSTENUTE:
    ym2151_set_hold( ev->ch,  ev->b>64?FLAG_TRUE:FLAG_FALSE );
    break;

  case SMF_CTRL_SOFT_PEDAL:
    break;

  case SMF_CTRL_HOLD2:
    break;


  case SMF_CTRL_REVERB:
    break;

  case SMF_CTRL_TREMOLO:
    break;

  case SMF_CTRL_CHORUS:
    break;

  case SMF_CTRL_DELAY:
    break;

  case SMF_CTRL_PHASER:
    break;


  case SMF_CTRL_DATA_INCLEMENT:
    break;

  case SMF_CTRL_DATA_DECLIMENT:
    break;

  case SMF_CTRL_NRPM_L:
    nrpn_adr[ev->ch] &= 0xff00;
    nrpn_adr[ev->ch] |= ev->b;
    rpn_adr[ev->ch] = 0xffff;
    break;

  case SMF_CTRL_NRPN_M:
    nrpn_adr[ev->ch] &= 0x00ff;
    nrpn_adr[ev->ch] |= (ev->b<<8);
    rpn_adr[ev->ch] = 0xffff;
    break;

  case SMF_CTRL_RPN_L:
    rpn_adr[ev->ch] &= 0xff00;
    rpn_adr[ev->ch] |= ev->b;
    nrpn_adr[ev->ch] = 0xffff;
    break;

  case SMF_CTRL_RPN_M:
    rpn_adr[ev->ch] &= 0x00ff;
    rpn_adr[ev->ch] |= (ev->b<<8);
    nrpn_adr[ev->ch] = 0xffff;
    break;


  case SMF_CTRL_ALL_SOUND_OFF:
    ym2151_all_note_off(ev->ch);
    reset_ym2151();
    break;

  case SMF_CTRL_RESET_ALL_CTRL:
    reset_ym2151();
    ym2151_set_hold( ev->ch,  FLAG_FALSE );
    ym2151_set_expression( ev->ch, 127 );
    ym2151_set_bend( ev->ch, 8192 );
    for ( i=0 ; i<VFB_MAX_CHANNEL_NUMBER ; i++ ) {
      rpn_adr[i]  = 0xffff;
      nrpn_adr[i] = 0xffff;
    }
    break;

  case SMF_CTRL_LOCAL_CONTROL:
    break;

  case SMF_CTRL_ALL_NOTE_OFF:
    ym2151_all_note_off(ev->ch);
    break;

  case SMF_CTRL_OMNI_MODE_OFF:
    break;

  case SMF_CTRL_OMNI_MODE_ON:
    break;

  case SMF_CTRL_MONO_MODE_ON:
    break;

  case SMF_CTRL_POLI_MODE_ON:
    break;

    
  default:
#ifdef VFB_DEBUG
    fprintf(stdout, "CTRLCHG:  %02x %02x %02x\n", ev->ch, ev->a, ev->b);
#endif
    break;
  }
  return 0;
}

/* ------------------------------------------------------------------- */

static int system_exclusive( MidiEvent *ev ) {

  int i,j;
  int d;

  if ( ev->ex_buf[0] == 0x7e ) { /* Universal non-realtime */
    if ( ev->ex_buf[2] == 0x09 && ev->ex_buf[3] == 0x01 ) {
      /* GM-MODE on */
      reset_ym2151();
      ym2151_all_note_off(ev->ch);
      ym2151_set_hold( ev->ch,  FLAG_FALSE );
      ym2151_set_expression( ev->ch, 127 );
      ym2151_set_bend( ev->ch, 8192 );
      for ( i=0 ; i<VFB_MAX_CHANNEL_NUMBER ; i++ ) {
	rpn_adr[i]  = 0xffff;
	nrpn_adr[i] = 0xffff;
      }
#ifdef VFB_DEBUG
      fprintf(stderr,"SYSEX: GM-Mode on\n");
#endif
    }
  }
  else if ( ev->ex_buf[0] == 0x7f ) { /* Universal realtime */
    if ( ev->ex_buf[2] == 0x04 && ev->ex_buf[3] == 0x01 ) {
      /* master volume */
      ym2151_set_system_volume( ev->ex_buf[5] );
#ifdef VFB_DEBUG
      fprintf(stderr,"SYSEX: Master volume: %d\n",ev->ex_buf[5]);
#endif
    }
  }
  else if ( ev->ex_buf[0] == 0x7d ) { /* vfb01 native */
    if ( ev->ex_buf[1] == 0x0a ) {    /* voice setting */
      int e[36];
      int sum=0,c;
      i=0;
      j=2;
      while(i<35) {
	c = ev->ex_buf[j++];
	if ( c==0xf7 ) break;
	e[i++] = c;
	sum+=c;
      }
      if ( i == 36 && sum&0x7f == 0 &&
	   e[0] < VFB_MAX_TONE_NUMBER && evfb != NULL) {
	VOICE_DATA *v = &evfb->voice[e[0]];
	j=0;
	v->voice_number = e[j++];
	for ( i=0 ; i<4 ; i++ ) {
	  v->ar[i]   = e[j++];
	  v->d1r[i]  = e[j++];
	  v->d2r[i]  = e[j++];
	  v->rr[i]   = e[j++];
	  v->sl[i]   = e[j++];
	  v->tl[i]   = e[j++];
	  v->ks[i]   = e[j++];
	  v->mul[i]  = e[j++];
	  v->dt1[i]  = e[j++];
	  v->dt2[i]  = e[j++];
	  v->ame[i]  = e[j++];
	}
	v->con       = e[j++];
	v->fl        = e[j++];
	v->slot_mask = e[j];
#ifdef VFB_DEBUG	
	fprintf(stderr,"SYSEX: Voice setting: %d\n",v->voice_number);
#endif
      }
    }
  }


#ifdef VFB_DEBUG
  i=0;
  fprintf(stdout, "SYSEX:    ");
  while (i<SYSEX_BUF_SIZE) {
    d = ev->ex_buf[i++];
    fprintf( stdout,"%02x ", d );fflush(stdout);
    if ( d == 0xf7 ) {
      fprintf( stdout, "\n" );
      break;
    }
  }
#endif
  return 0;
}

/* ------------------------------------------------------------------- */

static void sigexit( int num ) {

  closeVirtualMidiDevice();

#ifdef _POSIX_PRIORITY_SCHEDULING
  sched_yield();
#endif
#ifdef _POSIX_MEMLOCK
  munlockall();
#endif

  fprintf(stderr,_("Signal caught : %d\n"), num);
  exit(1);
}

static void sigquit( int num ) {
  closeVirtualMidiDevice();

#ifdef _POSIX_PRIORITY_SCHEDULING
  sched_yield();
#endif
#ifdef _POSIX_MEMLOCK
  munlockall();
#endif

  exit(0);
}

static void sigreset( int num ) {
  reset_ym2151();
  return;
}

static const int signals[]={SIGHUP,SIGQUIT,SIGILL,SIGABRT,SIGFPE,
                              SIGBUS,SIGSEGV,SIGPIPE,SIGTERM,SIGINT, 0};

static void set_signals( void ) {

  int i;

  for ( i=0 ; signals[i]!=0 ; i++ )
    signal( signals[i], sigexit );

  signal( SIGINT,  sigquit );
  signal( SIGUSR1, sigreset );

  return;
}

static void priority_init( void ) {

#ifdef _POSIX_PRIORITY_SCHEDULING
  struct sched_param ptmp, *priority_param;
  int i;

  priority_param=&ptmp;
  i=sched_get_priority_max( SCHED_FIFO );
  priority_param->sched_priority = i/2; /* no means */
  sched_setscheduler( 0, SCHED_FIFO, priority_param );
#endif
#ifdef _POSIX_MEMLOCK
  mlockall(MCL_CURRENT);
#endif

  return;
}

