/*
  VFB-01*16 / Virtual FB-01 module

  Copyright 2000 by Daisuke Nagano <breeze.nagano@nifty.ne.jp>
  Jun.25.2000


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "vfb01.h"
#include "smf.h"
#include "pcm8.h"

/* ------------------------------------------------------------------- */

static int midiDevice=-1;

int openVirtualMidiDevice( const char *devicename ) {

  if ( devicename == NULL ) return 1;
  midiDevice = open( devicename, O_RDONLY );
  if ( midiDevice < 0 ) return 1;

  return 0;
}

int closeVirtualMidiDevice() {

  if ( midiDevice >= 0 ) {
    close( midiDevice );
    midiDevice = -1;
  }
  return 0;
}

static int getMidiData() {

  fd_set fd;
  struct timeval tv;
  unsigned char buf[16];
  int ret=-1;

  while (1) {
    FD_ZERO( &fd );
    FD_SET( midiDevice, &fd );
    tv.tv_sec  = 0;
    tv.tv_usec = PCM8_SYSTEM_RATE*1000;

    if ( select( midiDevice+1, &fd, NULL, NULL, &tv ) < 0 ) {
#ifdef HAVE_NANOSLEEP
      struct timespec t;
      t.tv_sec = 0;
      t.tv_nsec=(long)1000*1000*PCM8_SYSTEM_RATE;
      nanosleep(&t,NULL);
#endif
      continue;
    }
    
    if ( FD_ISSET( midiDevice, &fd ) ) {
      read( midiDevice, buf, 1 );
      ret = (int)buf[0];
      break;
    }
  }

  return ret;
}

/* ------------------------------------------------------------------- */

int getMidiEvent( MidiEvent *ev ) {

  int d,l;
  static int last_event = -1;
  int d1,d2;

  ev->type = -1;

  d = getMidiData();
  if ( d < 0 ) return 0;
  
check_last_event:
  if ( d < 0x80 ) { /* running status */
    if ( last_event < 0 ) return 0;
  } else {
    last_event = d;
  }

  ev->ch = last_event&0x0f;

  switch(last_event&0xf0) {
  case 0x80:
  case 0x90:
  case 0xa0:
  case 0xb0:
  case 0xe0:
    d1 = getMidiData();
    if ( d1<0x00 ) return 0;
    if ( d1>0x7f ) { d=d1 ; goto check_last_event; }
    d2 = getMidiData();
    if ( d2<0x00 ) return 0;
    if ( d2>0x7f ) { d=d2 ; goto check_last_event; }
    ev->a = d1;
    ev->b = d2;
    break;

  case 0xc0:
  case 0xd0:
    d1 = getMidiData();
    if ( d1<0x00 ) return 0;
    if ( d1>0x7f ) { d=d1 ; goto check_last_event; }
    ev->a = d1;
    break;

  case 0xf0:
    switch(last_event) {
    case 0xf0:
      l=0;
      do {
	d1 = getMidiData();
	if ( l<SYSEX_BUF_SIZE )
	  ev->ex_buf[l++] = d1;
      } while(d1!=0xf7);
      if ( l==SYSEX_BUF_SIZE ) return 0;
      break;

    case 0xf1:
    case 0xf3:
    case 0xf5: /* port change */
      /* ignore following byte */
      getMidiData();
      break;

    case 0xf2:
      /* ignore folowing two bytes */
      getMidiData();
      getMidiData();
      break;

    default:
      return 1;
    }
    break;

  default:
    break;
  }

  /* process each controls */

  switch( last_event&0xf0 ) {
  case 0x80:
    ev->type = MIDI_NOTEOFF;
    break;

  case 0x90:
    if ( ev->b == 0 )
      ev->type = MIDI_NOTEOFF;
    else
      ev->type = MIDI_NOTEON;
    break;

  case 0xa0:
    ev->type = MIDI_PRESSURE;
    break;

  case 0xb0:
    ev->type = MIDI_CONTROL;
    break;

  case 0xc0:
    ev->type = MIDI_PROGRAM;
    break;

  case 0xd0:
    ev->type = MIDI_CHANPRES;
    break;

  case 0xe0:
    ev->type = MIDI_PITCHB;
    break;

  case 0xf0:
    ev->type = MIDI_SYSEX;
    break;

  default:
    break;
  }

  return 1;
}

